package main;

import java.util.HashMap;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Iterator;
import java.util.ArrayList;
import java.io.*;
import java.util.Map.Entry;
import java.util.Date;
import java.util.Scanner;

public class Main 
{
	public static void main(String[] args)
	{
		if (args.length == 0)
		{
			_printUsage();
			System.exit(0);
		}
		
		String input_file = "", output_directory = "", start_time = "", end_time = "";
		int long_call_limit_in_seconds = 0;
		ArrayList<String> calls_to_ignore = new ArrayList<>();
		for (int i = 0; i < args.length; i++)
		{
			String arg = args[i];
			if (arg.contentEquals("-i") || arg.contentEquals("--input-file"))
			{
				input_file = args[i+1];
				i++;
			}
			else if (arg.contentEquals("-o") || arg.contentEquals("--output-directory"))
			{
				output_directory = args[i+1];
				i++;
			}
			else if (arg.contentEquals("-s") || arg.contentEquals("--start-time"))
			{
				start_time = args[i+1];
				i++;
			}
			else if (arg.contentEquals("-e") || arg.contentEquals("--end-time"))
			{
				end_time = args[i+1];
				i++;
			}
			else if (arg.contentEquals("-l" ) || arg.contentEquals("--long-calls"))
			{
				long_call_limit_in_seconds = Integer.valueOf(args[i+1]);
				i++;
			}
			else if (arg.contentEquals("-f") || arg.contentEquals("--filter-calls"))
			{
				String filter_string = args[i+1];
				if (filter_string.indexOf(',') > 0)
				{
					try
					{
						Scanner s = new Scanner(filter_string);
						s.useDelimiter(",");
						while (s.hasNext())
						{
							String call = s.next();
							calls_to_ignore.add(call);
						}
						s.close();
					}
					catch (Exception ex)
					{
						System.err.println("--ERROR-- Caught an exception trying to parse --filter-calls argument:");
						ex.printStackTrace(System.err);
					}
				}
				else
				{
					calls_to_ignore.add(filter_string);
				}
			}
			else if (arg.contentEquals("-h")|| arg.contentEquals("--help"))
			{
				_printUsage();
				System.exit(0);
			}
		}
		
		if (!_validateArguments(input_file, output_directory, start_time, end_time, long_call_limit_in_seconds))
		{
			System.err.println("Argument validation failed. See below for expected inputs:");
			System.err.println();
			_printUsage();
			System.exit(1);
		}
		
		//If we made it to here, validations all check out so fire up the parser
		Parser parser = new Parser();
		String truncated_file_path = output_directory + "/truncated-access-log.txt";
		String results_file = output_directory + "/analysis.csv";
		
		//Truncate the log file down to the time frame we're interested in to save some memory
		parser.truncate(input_file, truncated_file_path, start_time, end_time);
		
		//verify that the truncated file now exists
		File truncated_file = new File(truncated_file_path);
		if (!truncated_file.exists())
		{
			System.err.println("--ERROR-- Truncated access log file was not found, can't continue analysis");
			System.exit(1);
		}
		
		//Parse the truncated file
		parser.parse(truncated_file_path, start_time, long_call_limit_in_seconds, calls_to_ignore);
		
		//Analyze the data
		StringBuilder analysis_contents = _analyzeResults(parser);
		
		//Save it to a file
		try
		{
			PrintWriter out = new PrintWriter(new FileWriter(results_file));
			out.println("Namespace,Api,Method,# CacheHits,CacheMisses,AvgTime,SlowestTime,FastestTime,#Success,#Failed,Total#Calls,#SlowCalls(1s+),AvgSlow,Slow%,TotalTime,TotalSlowTime,TotalFastTime,AvgFast");
			out.println(analysis_contents.toString());
			out.close();
		}
		catch (Exception ex)
		{
			System.err.println("--ERROR-- Caught an exception trying to write the results file:");
			ex.printStackTrace(System.err);
		}
		
		//Do we need to save the long call info?
		if (long_call_limit_in_seconds > 0)
		{
			ArrayList<String> long_calls = parser.getLongCalls();
			String long_calls_file = output_directory + "/slow_calls.txt";
			try
			{
				PrintWriter long_out = new PrintWriter(new FileWriter(long_calls_file));
				for (String call : long_calls)
				{
					long_out.println(call);
				}
				long_out.close();
			}
			catch (Exception ex)
			{
				System.err.println("--ERROR-- Caught an exception while trying to save the long calls file:");
				ex.printStackTrace(System.err);
			}
		}
		
		//clean up the truncated log file
		truncated_file.delete();
		
		System.out.println();
		System.out.println();
		System.out.println("--INFO-- Done analyzing! See " + results_file + " for results.");
		if (long_call_limit_in_seconds > 0)
		{
			System.out.println("--INFO-- Additional information on slow calls has been saved to " + output_directory + "/slow_calls.txt");
		}
	}
	
	private static StringBuilder _analyzeResults(Parser parser)
	{
		StringBuilder results = new StringBuilder();
		
		try
		{
			HashMap<String, ArrayList<BigDecimal>> response_time_map = parser.getResponseTimesMap();
			HashMap<String, Integer> method_successes = parser.getMethodSuccesses();
			HashMap<String, Integer> method_failures = parser.getMethodFailures();
			HashMap<String, Integer> cache_hits = parser.getCacheHits();
			HashMap<String, Integer> cache_misses = parser.getCacheMisses();
			HashMap<Integer, Integer> status_codes = parser.getStatusCodesMap();
			HashMap<String, Integer> slow_calls = parser.getSlowCallsMap();
			
			Iterator<Entry<Integer, Integer>> status_iter = status_codes.entrySet().iterator();
			Iterator<Entry<String, Integer>> slow_iter = slow_calls.entrySet().iterator();
			int total_count = 0;
			System.out.println();
			System.out.println("---------------------------------------------------------");
			System.out.println("Status Code\t:Count");
			while (status_iter.hasNext())
			{
				Entry<Integer, Integer> entry = status_iter.next();
				System.out.println(entry.getKey() + "\t: " + entry.getValue());
				total_count += entry.getValue();
			}
			System.out.println();
			System.out.println("Total # of Calls: " + total_count);
			System.out.println("---------------------------------------------------------");
			System.out.println();
			
			System.out.println();
			System.out.println("---------------------------------------------------------");
			int slow_count = 0;
			System.out.println("Slow Call\tCount");
			while (slow_iter.hasNext())
			{
				Entry<String, Integer> entry = slow_iter.next();
				System.out.println(entry.getKey() + "\t" + entry.getValue());
				slow_count += entry.getValue();
			}
			System.out.println();
			System.out.println("Total # of Slow Calls: " + slow_count);
			System.out.println("---------------------------------------------------------");
			System.out.println();
			
			
			Iterator<Entry<String, ArrayList<BigDecimal>>> response_iter = response_time_map.entrySet().iterator();
			while (response_iter.hasNext())
			{
				BigDecimal max_response = null;//new BigDecimal(0);
				BigDecimal min_response = null; //new BigDecimal(0);
				BigDecimal response_sum = new BigDecimal(0);
				
				BigDecimal slow_cutoff = new BigDecimal(1);
				int slow_responses = 0;
				BigDecimal slow_times = new BigDecimal(0);
				
				Entry<String, ArrayList<BigDecimal>> entry = response_iter.next();
				String method = entry.getKey();
				ArrayList<BigDecimal> response_times = entry.getValue();
				for (BigDecimal response_time : response_times)
				{
					if (min_response == null)
						min_response = response_time;
					if (max_response == null)
						max_response = response_time;
					
					if (response_time.compareTo(max_response) > 0)
						max_response = response_time;
					if (response_time.compareTo(min_response) < 0)
						min_response = response_time;
					response_sum = response_sum.add(response_time);
					
					if (response_time.compareTo(slow_cutoff) > 0)
					{
						slow_responses++;
						slow_times = slow_times.add(response_time);
					}
				}
				
				
				int hits = 0;
				if (cache_hits.containsKey(method))
					hits = cache_hits.get(method);
				
				int misses = 0;
				if (cache_misses.containsKey(method))
					misses = cache_misses.get(method);
				
				BigDecimal average_response_time = response_sum.divide(new BigDecimal(response_times.size()), 3, RoundingMode.HALF_UP);

				int successes = 0;
				if (method_successes.containsKey(method))
					successes = method_successes.get(method);
				
				int failures = 0;
				if (method_failures.containsKey(method))
					failures = method_failures.get(method);
				
				
				BigDecimal slow_average;// = slow_times.divide(new BigDecimal(slow_responses), 3, RoundingMode.HALF_UP);
				if (slow_responses > 0)
					slow_average = slow_times.divide(new BigDecimal(slow_responses), 3, RoundingMode.HALF_UP);
				else
					slow_average = new BigDecimal(0);
				
				BigDecimal slow_responses_bd = new BigDecimal(slow_responses);
				BigDecimal total_calls_bd = new BigDecimal(successes + failures);
				BigDecimal slow_percent = slow_responses_bd.divide(total_calls_bd, 3, RoundingMode.HALF_UP);
				slow_percent = slow_percent.multiply(new BigDecimal(100));
				BigDecimal total_time = average_response_time.multiply(total_calls_bd);
				BigDecimal slow_average_bd = slow_average.multiply(new BigDecimal(slow_responses));
				BigDecimal total_time_fast = total_time.subtract(slow_average_bd);
				BigDecimal slow_diff = new BigDecimal(successes + failures).subtract(new BigDecimal(slow_responses));
				BigDecimal average_fast = new BigDecimal("0"); //total_time_fast.divide(slow_diff, 3, RoundingMode.HALF_UP);
				if (slow_diff.compareTo(new BigDecimal("0")) != 0)
						average_fast = total_time_fast.divide(slow_diff, 3, RoundingMode.HALF_UP);
				
				int separator1 = method.indexOf('.');
				int separator2 = method.indexOf('.', separator1+1);
				String namespace = method.substring(0, separator1);
				String api = method.substring(separator1+1, separator2);
				String call = method.substring(separator2+1);
				namespace = namespace.replaceAll("\\.", "");
				api = api.replaceAll("\\.", "");
				call = call.replaceAll("\\.", "");
				results.append(namespace);
				results.append(",");
				results.append(api);
				results.append(",");
				results.append(call);
				results.append(",");
				results.append(hits);
				results.append(",");
				results.append(misses);
				results.append(",");
				results.append(average_response_time.stripTrailingZeros().toPlainString());
				results.append(",");
				results.append(max_response.stripTrailingZeros().toPlainString());
				results.append(",");
				results.append(min_response.stripTrailingZeros().toPlainString());
				results.append(",");
				results.append(successes);
				results.append(",");
				results.append(failures);
				results.append(",");
				results.append((successes + failures));
				results.append(",");
				results.append(slow_responses);
				results.append(",");
				results.append(slow_average);
				results.append(",");
				results.append(slow_percent.stripTrailingZeros().toPlainString() + "%");
				results.append(",");
				results.append(total_time.stripTrailingZeros().toPlainString());
				results.append(",");
				results.append(slow_average_bd.stripTrailingZeros().toPlainString());
				results.append(",");
				results.append(total_time_fast.stripTrailingZeros().toPlainString());
				results.append(",");
				results.append(average_fast.stripTrailingZeros().toPlainString());
				results.append("\r\n");
			}
		}
		catch (Exception ex)
		{
			System.err.println("Exception caught in _analyzeResults:");
			ex.printStackTrace(System.err);
		}
		
		return results;
	}
	
	private static boolean _validateArguments(String input, String output, String start_time, String end_time, int long_call_time)
	{
		File input_test = new File(input);
		if ("".contentEquals(input))
		{
			System.err.println("--Error-- No input file specified");
			return false;
		}
		
		if (!input_test.exists())
		{
			System.err.println("--Error-- Input file doesn't exist. Path: " + input);
			return false;
		}
		
		if ("".contentEquals(output))
		{
			System.err.println("--Error-- No output directory specified");
			return false;
		}
		
		File output_test = new File(output);
		if (!output_test.isDirectory())
		{
			System.err.println("--Error-- Output directory is not a valid directory. Path: " + output);
			return false;
		}
		
		if ("".contentEquals(start_time))
		{
			System.err.println("--Error-- Starting timestamp was not specified");
			return false;
		}
		
		if ("".contentEquals(end_time))
		{
			System.err.println("--Error-- Ending timestamp was not specified");
			return false;
		}
		
		DateFormat format = new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss");
		Date start_as_date_format = null;
		Date end_as_date_format = null;
		try
		{
			start_as_date_format = format.parse(start_time);
		}
		catch (Exception ex)
		{
			System.err.println("--Error-- Starting timestamp doesn't match the expected format. See info below for correct format");
			return false;
		}
		
		try
		{
			end_as_date_format = format.parse(end_time);
		}
		catch (Exception ex)
		{
			System.err.println("--Error-- Ending timestamp doesn't match the expected format. See info below for correct format");
			return false;
		}
		
		if (start_as_date_format != null && end_as_date_format != null)
		{
			if (end_as_date_format.compareTo(start_as_date_format) < 0)
			{
				System.err.println("--Error-- Ending timestamp can't be before starting timestamp");
				return false;
			}
		}
		else
		{
			System.err.println("--Error-- Missing starting or ending timestamp");
			return false;
		}
		
		if (long_call_time < 0)
			return false;
		else if (long_call_time > 20)
		{
			System.out.println("--WARNING-- you've specified that you only want to save calls that have a response time longer than " + long_call_time + " seconds. Continuing, but make sure that's what you want.");
		}
		
		return true;
	}
	
	private static void _printUsage()
	{
		System.out.println("Usage:");
		System.out.println("\t-i --input-file\t path to the nginx log file (access.log) -- Required");
		System.out.println("\t-o --output-directory\t path where the analysis file will be saved -- Required");
		System.out.println("\t-s --start-time\t timestamp to start analyzing log entries -- Required");
		System.out.println("\t-e --end-time\t timestamp to stop analyzing log entries at -- Required");
		System.out.println("\t-l --long-calls\t saves a separate report of any calls over the given limit in seconds -- Optional");
		System.out.println("\t-f --filter-calls <comma separated list of calls to ignore>\t ignore this list of calls when analyzing data");
		System.out.println("\t-h --help\tPrint this message and exit");
		System.out.println();
		System.out.println();
		System.out.println("\tThe traffic log file must follow the format as specified on the nginx server that communicates with jussi (TODO add sample)");
		System.out.println("\tTimestamp formats must also follow a specific format: (DD/mmm/YYYY:HH:MM:SS) Example: 19/Oct/2020:05:00:00");
		System.out.println("\tNote that the traffic logs are usually in the GMT timezone.");
		System.out.println("\tThe -l or --long-calls option takes an integer that represents the number of seconds that can pass before a call is considered slow. Passing 3 would mean that any calls that took longer than 3 seconds to respond will be logged in a separate file.");
	}
}
