package main;

import java.io.*;
import java.util.Scanner;
import java.util.HashMap;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.ArrayList;

public class Parser
{
	//Map <Method to Time>
	private final HashMap<String, ArrayList<BigDecimal>> _method_to_response_time_map;
	//Map <Method to Map<Cache Hit, Cache Miss>>
	private final HashMap<String, Integer> _cache_hits;
	private final HashMap<String, Integer> _cache_misses;
	private final HashMap<String, Integer> _method_failures;
	private final HashMap<String, Integer> _method_successes;
	private final HashMap<Integer, Integer> _http_codes;
	private final HashMap<String, ArrayList<BigDecimal>> _sql_query_response_time_map;
	private final HashMap<String, Integer> _slow_calls_map;
	private final HashMap<String, ArrayList<BigDecimal>> _special_map;
	private final ArrayList<String> _long_calls;
	
	public Parser()
	{
		_method_to_response_time_map = new HashMap<>();
		_cache_hits = new HashMap<>();
		_cache_misses = new HashMap<>();
		_method_failures = new HashMap<>();
		_method_successes = new HashMap<>();
		_http_codes = new HashMap<>();
		_sql_query_response_time_map = new HashMap<>();
		_slow_calls_map = new HashMap<>();
		_special_map = new HashMap<>();
		_long_calls = new ArrayList<>();
	}
	
	public ArrayList<String> getLongCalls()
	{
		return _long_calls;
	}
	
	public HashMap<String, ArrayList<BigDecimal>> getSpecialMap()
	{
		return _special_map;
	}
	
	public HashMap<String, Integer> getSlowCallsMap()
	{
		return _slow_calls_map;
	}
	
	public HashMap<String, ArrayList<BigDecimal>> getSQLResponseTimesMap()
	{
		return _sql_query_response_time_map;
	}
	
	public HashMap<String, Integer> getMethodSuccesses()
	{
		return _method_successes;
	}
	
	public HashMap<String, Integer> getMethodFailures()
	{
		return _method_failures;
	}
	
	public HashMap<String, ArrayList<BigDecimal>> getResponseTimesMap()
	{
		return _method_to_response_time_map;
	}
	
	public HashMap<String, Integer> getCacheHits()
	{
		return _cache_hits;
	}
	
	public HashMap<String, Integer> getCacheMisses()
	{
		return _cache_misses;
	}
	
	public HashMap<Integer, Integer> getStatusCodesMap()
	{
		return _http_codes;
	}
	
	public void truncate(String input_file, String output_file, String start_date, String end_date)
	{
		try
		{
			//DateFormat format = new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss");
			//Date starting_date = format.parse("17/Apr/2020:21:41:00");
			
			BufferedReader reader = new BufferedReader(new FileReader(input_file));
			PrintWriter writer = new PrintWriter(new FileWriter(output_file));
			String line = "";
			DateFormat format = new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss");
			Date starting_time = format.parse(start_date);
			Date ending_time = format.parse(end_date);
			System.out.println("--INFO-- Truncating input file down to the desired time range. This may take a few minutes.");
			while ((line = reader.readLine()) != null)
			{
				if (line.contains("||"))
				{
					try
					{
						Scanner ls = new Scanner(line);
						ls.useDelimiter("\\|\\|");
						ls.next();
						String time = ls.next();
						time = time.trim();
						time = time.substring(1, time.length() - 7);
						ls.close();
						Date line_date = format.parse(time);
						if (starting_time.compareTo(line_date) > 0)
							continue;
						if (ending_time.compareTo(line_date) < 0)
							continue;
						writer.println(line);
					}
					catch (Exception ex)
					{
						continue;
					}
				}
				else
				{
					continue;
				}
			}
			
			reader.close();
			writer.close();
			System.out.println("--INFO-- Done truncating input file");
		}
		catch (Exception ex)
		{
			System.err.println("Parser.truncate caught an exception:");
			ex.printStackTrace(System.err);
		}
	}
	
	public void parse(String filename, String start_time, int long_call_limit_in_seconds, ArrayList<String> call_filter)
	{
		System.out.println("--INFO-- Starting parsing");
		int records_parsed = 0;
		try
		{
			File file = new File(filename);
			if (!file.exists())
				throw new FileNotFoundException("no such file: " + filename);
			
			DateFormat format = new SimpleDateFormat("dd/MMM/yyyy:HH:mm:ss");
			Date starting_date = format.parse(start_time);
			
			BufferedReader reader = new BufferedReader(new FileReader(file));
			String line = "";
			while ((line = reader.readLine()) != null)
			{
				Scanner ls = new Scanner(line);
				ls.useDelimiter("\\|\\|");
				ls.next();
				String time = ls.next();
				time = time.trim();
				time = time.substring(1, time.length() - 7);
				Date line_date = format.parse(time);
				if (starting_date.compareTo(line_date) > 0)
					continue;
				ls.next(); //http method
				int status_code = Integer.valueOf(ls.next().trim());
				ls.next(); //bytes sent
				ls.next(); //referrer
				ls.next(); // user agent
				ls.next(); //forwarded for
				String cache_hit = ls.next().trim();
				boolean was_cache_hit = !cache_hit.equals("\"-\"");
				String namespace = ls.next().trim(); //jussi name space
				String api = ls.next().trim();
				String method = namespace + "." + api + "." + ls.next().trim();
				String parameters = ls.next().trim(); //parameters
				ls.next(); //skip upstream connect time
				double upstream_response_time = 0.0;// = Double.valueOf(ls.next().trim());
				try
				{
					upstream_response_time = Double.valueOf(ls.next().trim());
				}
				catch(Exception ex)
				{
					continue;
				}
				ls.close();
				
				boolean should_skip = false;
				for (String filter : call_filter)
				{
					if (method.contains(filter))
					{
						should_skip = true;
						break;
					}
				}
				
				if (should_skip)
					continue;
				
				if (method.contains("get_account_posts") && parameters.contains("replies"))
				{
					ArrayList<BigDecimal> special_times = _special_map.get(method);
					if (special_times == null)
						special_times = new ArrayList<>();
					special_times.add(new BigDecimal(upstream_response_time));
					_special_map.put(method, special_times);
				}
				
				ArrayList<BigDecimal> response_times = _method_to_response_time_map.get(method);
				if (response_times == null)
					response_times = new ArrayList<BigDecimal>();
				response_times.add(new BigDecimal(upstream_response_time));
				if (long_call_limit_in_seconds > 0 && upstream_response_time > long_call_limit_in_seconds)
				{
					int count = 0;
					if (_slow_calls_map.containsKey(method))
						count = _slow_calls_map.get(method);
					count++;
					_slow_calls_map.put(method, count);
					
					String full_call = time + "\t" + method + "\t" + parameters + "\t\t" + upstream_response_time;
					_long_calls.add(full_call);
				}
				_method_to_response_time_map.put(method, response_times);
				if (was_cache_hit)
				{
					int count = 0;
					if (_cache_hits.containsKey(method))
						count = _cache_hits.get(method);
					count++;
					_cache_hits.put(method, count);
				}
				else
				{
					int count = 0;
					if (_cache_misses.containsKey(method))
						count = _cache_misses.get(method);
					count++;
					_cache_misses.put(method, count);
				}
				
				if (status_code < 400)
				{
					//200 or a 300 something, indicating probable success
					int count = 0;
					if (_method_successes.containsKey(method))
						count = _method_successes.get(method);
					count++;
					_method_successes.put(method, count);
				}
				else
				{
					//400 or 500's are definitely errors
					int count = 0;
					if (_method_failures.containsKey(method))
						count = _method_failures.get(method);
					count++;
					_method_failures.put(method, count);
				}
				
				int status_count = 0;
				if (_http_codes.get(status_code) != null)
					status_count = _http_codes.get(status_code);
				status_count++;
				_http_codes.put(status_code, status_count);
				
				records_parsed++;
				if (records_parsed % 500 == 0)
					System.out.println("--INFO--\t Parsed " + records_parsed + " records so far.");
			}
			reader.close();
			//out.close();
			System.out.println("--INFO-- Done parsing");
		}
		catch (Exception ex)
		{
			System.err.println("Parser.parse caught an exception:");
			ex.printStackTrace(System.err);
		}
	}
	
	void parseSQLTimes(String filepath)
	{
		int line_number = 0;
		try
		{
			File file = new File(filepath);
			if (!file.exists())
				throw new FileNotFoundException("No such file: " + filepath);
			
			BufferedReader reader = new BufferedReader(new FileReader(file));
			String line;
			while ((line = reader.readLine()) != null)
			{
				line_number++;
				if (!line.contains("---"))
					continue;
				Scanner ls = new Scanner(line);
				ls.useDelimiter("\\[");
				ls.next();
				String timing_string = ls.next();
				String time = timing_string.substring(0, timing_string.lastIndexOf("]") - 2);
				String rol = timing_string.substring(timing_string.lastIndexOf("]"));
				int pos = rol.indexOf("---");
				if (pos < 0)
					continue;
				int pos2 = rol.indexOf(" ", pos);
				if (pos2 < 0)
					continue;
				String query_method = rol.substring(pos+3, pos2);
				
				BigDecimal response_time = new BigDecimal(time);
				response_time = response_time.divide(new BigDecimal("1000"), 3, RoundingMode.HALF_EVEN);
				ArrayList<BigDecimal> times = _sql_query_response_time_map.get(query_method);
				if (times == null)
					times = new ArrayList<>();
				times.add(response_time);
				_sql_query_response_time_map.put(query_method, times);
				ls.close();
			}
			reader.close();
		}
		catch (Exception ex)
		{
			System.err.println(ex.toString());
			ex.printStackTrace(System.err);
			System.err.println("Last line processed: " + line_number);
		}
	}
}
